@@ -2,7 +2,7 @@
 
   ruby.c -
 
-  $Author: nobu $
+  $Author: naruse $
   created at: Tue Aug 10 12:47:31 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -15,8 +15,10 @@
 #include <windows.h>
 #include <sys/cygwin.h>
 #endif
-#include "internal.h"
+#include "ruby/encoding.h"
 #include "ruby/thread.h"
+#include "ruby/version.h"
+#include "internal.h"
 #include "eval_intern.h"
 #include "dln.h"
 #include <stdio.h>
@@ -50,6 +52,10 @@
 
 #include "ruby/util.h"
 
+#include "mjit.h"
+
+void Init_ruby_description(void);
+
 #ifndef HAVE_STDLIB_H
 char *getenv();
 #endif
@@ -73,6 +79,8 @@ char *getenv();
     X(rubyopt) \
     SEP \
     X(frozen_string_literal) \
+    SEP \
+    X(jit) \
     /* END OF FEATURES */
 #define EACH_DEBUG_FEATURES(X, SEP) \
     X(frozen_string_literal) \
@@ -134,6 +142,9 @@ struct ruby_cmdline_options {
     VALUE req_list;
     unsigned int features;
     unsigned int dump;
+#if USE_MJIT
+    struct mjit_options mjit;
+#endif
     int safe_level;
     int sflag, xflag;
     unsigned int warning: 1;
@@ -162,6 +173,7 @@ enum {
 	& ~FEATURE_BIT(gems)
 #endif
 	& ~FEATURE_BIT(frozen_string_literal)
+        & ~FEATURE_BIT(jit)
 	)
 };
 
@@ -174,6 +186,9 @@ cmdline_options_init(ruby_cmdline_options_t *opt)
     opt->ext.enc.index = -1;
     opt->intern.enc.index = -1;
     opt->features = DEFAULT_FEATURES;
+#ifdef MJIT_FORCE_ENABLE /* to use with: ./configure cppflags="-DMJIT_FORCE_ENABLE" */
+    opt->features |= FEATURE_BIT(jit);
+#endif
     return opt;
 }
 
@@ -192,7 +207,7 @@ static void
 show_usage_line(const char *str, unsigned int namelen, unsigned int secondlen, int help)
 {
     const unsigned int w = 16;
-    const int wrap = help && namelen + secondlen - 2 > w;
+    const int wrap = help && namelen + secondlen - 1 > w;
     printf("  %.*s%-*.*s%-*s%s\n", namelen-1, str,
 	   (wrap ? 0 : w - namelen + 1),
 	   (help ? secondlen-1 : 0), str + namelen,
@@ -233,10 +248,12 @@ usage(const char *name, int help)
 	M("-s",		   "",			   "enable some switch parsing for switches after script name"),
 	M("-S",		   "",			   "look for the script using PATH environment variable"),
 	M("-T[level=1]",   "",			   "turn on tainting checks"),
-	M("-v",		   ", --verbose",	   "print version number, then turn on verbose mode"),
+	M("-v",		   "",			   "print the version number, then turn on verbose mode"),
 	M("-w",		   "",			   "turn warnings on for your script"),
 	M("-W[level=2]",   "",			   "set warning level; 0=silence, 1=medium, 2=verbose"),
 	M("-x[directory]", "",			   "strip off text before #!ruby line and perhaps cd to directory"),
+        M("--jit",         "",                     "enable JIT with default options (experimental)"),
+        M("--jit-[option]","",                     "enable JIT with an option (experimental)"),
 	M("-h",		   "",			   "show this message, --help for more info"),
     };
     static const struct message help_msg[] = {
@@ -247,7 +264,8 @@ usage(const char *name, int help)
 	  "enable or disable features. see below for available features"),
 	M("--external-encoding=encoding",           ", --internal-encoding=encoding",
 	  "specify the default external or internal character encoding"),
-	M("--version",                              "", "print the version"),
+	M("--verbose",                              "", "turn on verbose mode and disable script from stdin"),
+	M("--version",                              "", "print the version number, then exit"),
 	M("--help",			            "", "show this message, -h for short message"),
     };
     static const struct message dumps[] = {
@@ -261,6 +279,16 @@ usage(const char *name, int help)
 	M("did_you_mean", "",   "did_you_mean (default: "DEFAULT_RUBYGEMS_ENABLED")"),
 	M("rubyopt", "",        "RUBYOPT environment variable (default: enabled)"),
 	M("frozen-string-literal", "", "freeze all string literals (default: disabled)"),
+        M("jit", "",            "JIT compiler (default: disabled)"),
+    };
+    static const struct message mjit_options[] = {
+        M("--jit-warnings",      "", "Enable printing JIT warnings"),
+        M("--jit-debug",         "", "Enable JIT debugging (very slow)"),
+        M("--jit-wait",          "", "Wait until JIT compilation is finished everytime (for testing)"),
+        M("--jit-save-temps",    "", "Save JIT temporary files in $TMP or /tmp (for testing)"),
+        M("--jit-verbose=num",   "", "Print JIT logs of level num or less to stderr (default: 0)"),
+        M("--jit-max-cache=num", "", "Max number of methods to be JIT-ed in a cache (default: 1000)"),
+        M("--jit-min-calls=num", "", "Number of calls to trigger JIT (for testing, default: 5)"),
     };
     int i;
     const int num = numberof(usage_msg) - (help ? 1 : 0);
@@ -280,6 +308,9 @@ usage(const char *name, int help)
     puts("Features:");
     for (i = 0; i < numberof(features); ++i)
 	SHOW(features[i]);
+    puts("JIT options (experimental):");
+    for (i = 0; i < numberof(mjit_options); ++i)
+	SHOW(mjit_options[i]);
 }
 
 #define rubylib_path_new rb_str_new
@@ -450,14 +481,59 @@ ruby_init_loadpath(void)
     ruby_init_loadpath_safe(0);
 }
 
-#if defined(LOAD_RELATIVE) && defined(HAVE_DLADDR) && !defined(__CYGWIN__)
+#if defined(LOAD_RELATIVE)
 static VALUE
-dladdr_path(const void* addr)
+runtime_libruby_path(void)
 {
+#if defined _WIN32 || defined __CYGWIN__
+    DWORD len = RSTRING_EMBED_LEN_MAX, ret;
+    VALUE path;
+    VALUE wsopath = rb_str_new(0, len*sizeof(WCHAR));
+    WCHAR *wlibpath;
+    char *libpath;
+
+    while (wlibpath = (WCHAR *)RSTRING_PTR(wsopath),
+	   ret = GetModuleFileNameW(libruby, wlibpath, len),
+	   (ret == len))
+    {
+	rb_str_modify_expand(wsopath, len*sizeof(WCHAR));
+	rb_str_set_len(wsopath, (len += len)*sizeof(WCHAR));
+    }
+    if (!ret || ret > len) rb_fatal("failed to get module file name");
+#if defined __CYGWIN__
+    {
+	const int win_to_posix = CCP_WIN_W_TO_POSIX | CCP_RELATIVE;
+	size_t newsize = cygwin_conv_path(win_to_posix, wlibpath, 0, 0);
+	if (!newsize) rb_fatal("failed to convert module path to cygwin");
+	path = rb_str_new(0, newsize);
+	libpath = RSTRING_PTR(path);
+	if (cygwin_conv_path(win_to_posix, wlibpath, libpath, newsize)) {
+	    rb_str_resize(path, 0);
+	}
+    }
+#else
+    {
+	DWORD i;
+	for (len = ret, i = 0; i < len; ++i) {
+	    if (wlibpath[i] == L'\\') {
+		wlibpath[i] = L'/';
+		ret = i+1;	/* chop after the last separator */
+	    }
+	}
+    }
+    len = WideCharToMultiByte(CP_UTF8, 0, wlibpath, ret, NULL, 0, NULL, NULL);
+    path = rb_utf8_str_new(0, len);
+    libpath = RSTRING_PTR(path);
+    WideCharToMultiByte(CP_UTF8, 0, wlibpath, ret, libpath, len, NULL, NULL);
+#endif
+    rb_str_resize(wsopath, 0);
+    return path;
+#elif defined(HAVE_DLADDR)
     Dl_info dli;
     VALUE fname, path;
+    const void* addr = (void *)(VALUE)expand_include_path;
 
-    if (!dladdr(addr, &dli)) {
+    if (!dladdr((void *)addr, &dli)) {
 	return rb_str_new(0, 0);
     }
 #ifdef __linux__
@@ -472,102 +548,66 @@ dladdr_path(const void* addr)
     }
     rb_str_resize(fname, 0);
     return path;
+#else
+# error relative load path is not supported on this platform.
+#endif
 }
 #endif
 
 #define INITIAL_LOAD_PATH_MARK rb_intern_const("@gem_prelude_index")
 
+VALUE ruby_archlibdir_path, ruby_prefix_path;
+
 void
 ruby_init_loadpath_safe(int safe_level)
 {
-    VALUE load_path;
+    VALUE load_path, archlibdir = 0;
     ID id_initial_load_path_mark;
     const char *paths = ruby_initial_load_paths;
 #if defined LOAD_RELATIVE
-# if defined HAVE_DLADDR || defined __CYGWIN__ || defined _WIN32
-#   define VARIABLE_LIBPATH 1
-# else
-#   define VARIABLE_LIBPATH 0
-# endif
-# if VARIABLE_LIBPATH
+#if !defined ENABLE_MULTIARCH
+# define RUBY_ARCH_PATH ""
+#elif defined RUBY_ARCH
+# define RUBY_ARCH_PATH "/"RUBY_ARCH
+#else
+# define RUBY_ARCH_PATH "/"RUBY_PLATFORM
+#endif
     char *libpath;
     VALUE sopath;
-# else
-    char libpath[MAXPATHLEN + 1];
-# endif
     size_t baselen;
-    char *p;
+    const char *p;
 
-#if defined _WIN32 || defined __CYGWIN__
-    {
-	DWORD len = RSTRING_EMBED_LEN_MAX, ret, i;
-	VALUE wsopath = rb_str_new(0, len*sizeof(WCHAR));
-	WCHAR *wlibpath;
-	while (wlibpath = (WCHAR *)RSTRING_PTR(wsopath),
-	       ret = GetModuleFileNameW(libruby, wlibpath, len),
-	       (ret == len))
-	{
-	    rb_str_modify_expand(wsopath, len*sizeof(WCHAR));
-	    rb_str_set_len(wsopath, (len += len)*sizeof(WCHAR));
-	}
-	if (!ret || ret > len) rb_fatal("failed to get module file name");
-	for (len = ret, i = 0; i < len; ++i) {
-	    if (wlibpath[i] == L'\\') {
-		wlibpath[i] = L'/';
-		ret = i+1;	/* chop after the last separator */
-	    }
-	}
-	len = WideCharToMultiByte(CP_UTF8, 0, wlibpath, ret, NULL, 0, NULL, NULL);
-	sopath = rb_utf8_str_new(0, len);
-	libpath = RSTRING_PTR(sopath);
-	WideCharToMultiByte(CP_UTF8, 0, wlibpath, ret, libpath, len, NULL, NULL);
-	rb_str_resize(wsopath, 0);
-    }
-#elif defined(HAVE_DLADDR)
-    sopath = dladdr_path((void *)(VALUE)expand_include_path);
+    sopath = runtime_libruby_path();
     libpath = RSTRING_PTR(sopath);
-#endif
 
-#if !VARIABLE_LIBPATH
-    libpath[sizeof(libpath) - 1] = '\0';
-#endif
-#if defined DOSISH && !defined _WIN32
-    translit_char(libpath, '\\', '/');
-#elif defined __CYGWIN__
-    {
-	const int win_to_posix = CCP_WIN_A_TO_POSIX | CCP_RELATIVE;
-	size_t newsize = cygwin_conv_path(win_to_posix, libpath, 0, 0);
-	if (newsize > 0) {
-	    VALUE rubylib = rb_str_new(0, newsize);
-	    p = RSTRING_PTR(rubylib);
-	    if (cygwin_conv_path(win_to_posix, libpath, p, newsize) == 0) {
-		rb_str_resize(sopath, 0);
-		sopath = rubylib;
-		libpath = p;
-	    }
-	}
-    }
-#endif
     p = strrchr(libpath, '/');
     if (p) {
-	static const char bindir[] = "/bin";
+	static const char libdir[] = "/"
 #ifdef LIBDIR_BASENAME
-	static const char libdir[] = "/"LIBDIR_BASENAME;
+	    LIBDIR_BASENAME
 #else
-	static const char libdir[] = "/lib";
+	    "lib"
 #endif
+	    RUBY_ARCH_PATH;
+	const ptrdiff_t libdir_len = (ptrdiff_t)sizeof(libdir)
+	    - rb_strlen_lit(RUBY_ARCH_PATH) - 1;
+	static const char bindir[] = "/bin";
 	const ptrdiff_t bindir_len = (ptrdiff_t)sizeof(bindir) - 1;
-	const ptrdiff_t libdir_len = (ptrdiff_t)sizeof(libdir) - 1;
+
+	const char *p2 = NULL;
 
 #ifdef ENABLE_MULTIARCH
-	char *p2 = NULL;
-
       multiarch:
 #endif
 	if (p - libpath >= bindir_len && !STRNCASECMP(p - bindir_len, bindir, bindir_len)) {
 	    p -= bindir_len;
+	    archlibdir = rb_str_subseq(sopath, 0, p - libpath);
+	    rb_str_cat_cstr(archlibdir, libdir);
+	    OBJ_FREEZE_RAW(archlibdir);
 	}
 	else if (p - libpath >= libdir_len && !strncmp(p - libdir_len, libdir, libdir_len)) {
+	    archlibdir = rb_str_subseq(sopath, 0, (p2 ? p2 : p) - libpath);
+	    OBJ_FREEZE_RAW(archlibdir);
 	    p -= libdir_len;
 	}
 #ifdef ENABLE_MULTIARCH
@@ -580,33 +620,26 @@ ruby_init_loadpath_safe(int safe_level)
 	    if (p) goto multiarch;
 	    p = p2;
 	}
-#endif
-#if !VARIABLE_LIBPATH
-	*p = 0;
 #endif
     }
-#if !VARIABLE_LIBPATH
-    else {
-	strlcpy(libpath, ".", sizeof(libpath));
-	p = libpath + 1;
-    }
-    baselen = p - libpath;
-#define PREFIX_PATH() rb_str_new(libpath, baselen)
-#else
     baselen = p - libpath;
     rb_str_resize(sopath, baselen);
     libpath = RSTRING_PTR(sopath);
 #define PREFIX_PATH() sopath
-#endif
-
 #define BASEPATH() rb_str_buf_cat(rb_str_buf_new(baselen+len), libpath, baselen)
-
 #define RUBY_RELATIVE(path, len) rb_str_buf_cat(BASEPATH(), (path), (len))
 #else
     const size_t exec_prefix_len = strlen(ruby_exec_prefix);
 #define RUBY_RELATIVE(path, len) rubylib_path_new((path), (len))
 #define PREFIX_PATH() RUBY_RELATIVE(ruby_exec_prefix, exec_prefix_len)
 #endif
+    rb_gc_register_address(&ruby_prefix_path);
+    ruby_prefix_path = PREFIX_PATH();
+    OBJ_FREEZE_RAW(ruby_prefix_path);
+    if (!archlibdir) archlibdir = ruby_prefix_path;
+    rb_gc_register_address(&ruby_archlibdir_path);
+    ruby_archlibdir_path = archlibdir;
+
     load_path = GET_VM()->load_path;
 
     if (safe_level == 0) {
@@ -622,7 +655,7 @@ ruby_init_loadpath_safe(int safe_level)
 	paths += len + 1;
     }
 
-    rb_const_set(rb_cObject, rb_intern_const("TMP_RUBY_PREFIX"), rb_obj_freeze(PREFIX_PATH()));
+    rb_const_set(rb_cObject, rb_intern_const("TMP_RUBY_PREFIX"), ruby_prefix_path);
 }
 
 
@@ -633,11 +666,9 @@ add_modules(VALUE *req_list, const char *mod)
     VALUE feature;
 
     if (!list) {
-	*req_list = list = rb_ary_new();
-	RBASIC_CLEAR_CLASS(list);
+	*req_list = list = rb_ary_tmp_new(0);
     }
-    feature = rb_str_new2(mod);
-    RBASIC_CLEAR_CLASS(feature);
+    feature = rb_str_cat_cstr(rb_str_tmp_new(0), mod);
     rb_ary_push(list, feature);
 }
 
@@ -739,6 +770,7 @@ moreswitches(const char *s, ruby_cmdline_options_t *opt, int envopt)
     char **argv, *p;
     const char *ap = 0;
     VALUE argstr, argary;
+    void *ptr;
 
     while (ISSPACE(*s)) s++;
     if (!*s) return;
@@ -761,7 +793,8 @@ moreswitches(const char *s, ruby_cmdline_options_t *opt, int envopt)
     argc = RSTRING_LEN(argary) / sizeof(ap);
     ap = 0;
     rb_str_cat(argary, (char *)&ap, sizeof(ap));
-    argv = (char **)RSTRING_PTR(argary);
+    argv = ptr = ALLOC_N(char *, argc);
+    MEMMOVE(argv, RSTRING_PTR(argary), char *, argc);
 
     while ((i = proc_options(argc, argv, opt, envopt)) > 1 && envopt && (argc -= i) > 0) {
 	argv += i;
@@ -774,6 +807,7 @@ moreswitches(const char *s, ruby_cmdline_options_t *opt, int envopt)
 	}
     }
 
+    ruby_xfree(ptr);
     /* get rid of GC */
     rb_str_resize(argary, 0);
     rb_str_resize(argstr, 0);
@@ -916,6 +950,39 @@ set_option_encoding_once(const char *type, VALUE *name, const char *e, long elen
 #define set_source_encoding_once(opt, e, elen) \
     set_option_encoding_once("source", &(opt)->src.enc.name, (e), (elen))
 
+#if USE_MJIT
+static void
+setup_mjit_options(const char *s, struct mjit_options *mjit_opt)
+{
+    if (*s == 0) return;
+    else if (strcmp(s, "-warnings") == 0) {
+        mjit_opt->warnings = 1;
+    }
+    else if (strcmp(s, "-debug") == 0) {
+        mjit_opt->debug = 1;
+    }
+    else if (strcmp(s, "-wait") == 0) {
+        mjit_opt->wait = 1;
+    }
+    else if (strcmp(s, "-save-temps") == 0) {
+        mjit_opt->save_temps = 1;
+    }
+    else if (strncmp(s, "-verbose=", 9) == 0) {
+        mjit_opt->verbose = atoi(s + 9);
+    }
+    else if (strncmp(s, "-max-cache=", 11) == 0) {
+        mjit_opt->max_cache_size = atoi(s + 11);
+    }
+    else if (strncmp(s, "-min-calls=", 11) == 0) {
+        mjit_opt->min_calls = atoi(s + 11);
+    }
+    else {
+        rb_raise(rb_eRuntimeError,
+                 "invalid MJIT option `%s' (--help will show valid MJIT options)", s + 1);
+    }
+}
+#endif
+
 static long
 proc_options(long argc, char **argv, ruby_cmdline_options_t *opt, int envopt)
 {
@@ -1268,6 +1335,14 @@ proc_options(long argc, char **argv, ruby_cmdline_options_t *opt, int envopt)
 		opt->verbose = 1;
 		ruby_verbose = Qtrue;
 	    }
+            else if (strncmp("jit", s, 3) == 0) {
+#if USE_MJIT
+                opt->features |= FEATURE_BIT(jit);
+                setup_mjit_options(s + 3, &opt->mjit);
+#else
+                rb_warn("MJIT support is disabled.");
+#endif
+            }
 	    else if (strcmp("yydebug", s) == 0) {
 		if (envopt) goto noenvopt_long;
 		opt->dump |= DUMP_BIT(yydebug);
@@ -1354,6 +1429,7 @@ opt_enc_index(VALUE enc_name)
 #define rb_progname      (GET_VM()->progname)
 #define rb_orig_progname (GET_VM()->orig_progname)
 VALUE rb_argv0;
+VALUE rb_e_script;
 
 static VALUE
 false_value(void)
@@ -1507,7 +1583,15 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
     if (opt->src.enc.name)
 	rb_warning("-K is specified; it is for 1.8 compatibility and may cause odd behavior");
 
+#if USE_MJIT
+    if (opt->features & FEATURE_BIT(jit)) {
+        opt->mjit.on = TRUE; /* set mjit.on for ruby_show_version() API and check to call mjit_init() */
+    }
+#endif
     if (opt->dump & (DUMP_BIT(version) | DUMP_BIT(version_v))) {
+#if USE_MJIT
+        mjit_opts.on = opt->mjit.on; /* used by ruby_show_version(). mjit_init() still can't be called here. */
+#endif
 	ruby_show_version();
 	if (opt->dump & DUMP_BIT(version)) return Qtrue;
     }
@@ -1559,6 +1643,14 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
 
     ruby_gc_set_params(opt->safe_level);
     ruby_init_loadpath_safe(opt->safe_level);
+
+#if USE_MJIT
+    if (opt->mjit.on)
+        /* Using TMP_RUBY_PREFIX created by ruby_init_loadpath_safe(). */
+        mjit_init(&opt->mjit);
+#endif
+
+    Init_ruby_description();
     Init_enc();
     lenc = rb_locale_encoding();
     rb_enc_associate(rb_progname, lenc);
@@ -1623,9 +1715,9 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
     Init_ext();		/* load statically linked extensions before rubygems */
     if (opt->features & FEATURE_BIT(gems)) {
 	rb_define_module("Gem");
-    }
-    if (opt->features & FEATURE_BIT(did_you_mean)) {
-	rb_define_module("DidYouMean");
+	if (opt->features & FEATURE_BIT(did_you_mean)) {
+	    rb_define_module("DidYouMean");
+	}
     }
     ruby_init_prelude();
     if ((opt->features ^ DEFAULT_FEATURES) & COMPILATION_FEATURES) {
@@ -1704,7 +1796,7 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
 	rb_enc_set_default_internal(Qnil);
     rb_stdio_set_default_encoding();
 
-    if (!ast->root) {
+    if (!ast->body.root) {
 	rb_ast_dispose(ast);
 	return Qfalse;
     }
@@ -1726,7 +1818,7 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
     }
 
     if (dump & (DUMP_BIT(parsetree)|DUMP_BIT(parsetree_with_comment))) {
-	rb_io_write(rb_stdout, rb_parser_dump_tree(ast->root, dump & DUMP_BIT(parsetree_with_comment)));
+	rb_io_write(rb_stdout, rb_parser_dump_tree(ast->body.root, dump & DUMP_BIT(parsetree_with_comment)));
 	rb_io_flush(rb_stdout);
 	dump &= ~DUMP_BIT(parsetree)&~DUMP_BIT(parsetree_with_comment);
 	if (!dump) {
@@ -1749,7 +1841,7 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
 	    }
 	}
 	base_block = toplevel_context(toplevel_binding);
-	iseq = rb_iseq_new_main(ast->root, opt->script_name, path, vm_block_iseq(base_block));
+	iseq = rb_iseq_new_main(&ast->body, opt->script_name, path, vm_block_iseq(base_block));
 	rb_ast_dispose(ast);
     }
 
@@ -1765,8 +1857,24 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
     rb_define_readonly_boolean("$-l", opt->do_line);
     rb_define_readonly_boolean("$-a", opt->do_split);
 
+    if ((rb_e_script = opt->e_script) != 0) {
+        rb_gc_register_mark_object(opt->e_script);
+    }
+
     rb_set_safe_level(opt->safe_level);
 
+    {
+        rb_execution_context_t *ec = GET_EC();
+
+        if (opt->e_script) {
+            /* -e */
+            rb_exec_event_hook_script_compiled(ec, iseq, opt->e_script);
+        }
+        else {
+            /* file */
+            rb_exec_event_hook_script_compiled(ec, iseq, Qnil);
+        }
+    }
     return (VALUE)iseq;
 }
 
@@ -1836,11 +1944,7 @@ load_file_internal(VALUE argp_v)
 	c = rb_io_getbyte(f);
 	if (c == INT2FIX('#')) {
 	    c = rb_io_getbyte(f);
-	    if (c == INT2FIX('!')) {
-		line = rb_io_gets(f);
-		if (NIL_P(line))
-		    return 0;
-
+            if (c == INT2FIX('!') && !NIL_P(line = rb_io_gets(f))) {
 		RSTRING_GETMEM(line, str, len);
 		warn_cr_in_shebang(str, len);
 		if ((p = strstr(str, ruby_engine)) == 0) {
@@ -1964,7 +2068,7 @@ open_load_file(VALUE fname_v, int *xflag)
 #endif
 
 	if ((fd = rb_cloexec_open(fname, mode, 0)) < 0) {
-	    int e = errno;
+	    e = errno;
 	    if (!rb_gc_for_fd(e)) {
 		rb_load_fail(fname_v, strerror(e));
 	    }
@@ -2112,7 +2216,9 @@ external_str_new_cstr(const char *p)
 {
 #if UTF8_PATH
     VALUE str = rb_utf8_str_new_cstr(p);
-    return str_conv_enc(str, NULL, rb_default_external_encoding());
+    str = str_conv_enc(str, NULL, rb_default_external_encoding());
+    OBJ_TAINT_RAW(str);
+    return str;
 #else
     return rb_external_str_new_cstr(p);
 #endif
